// vgm_ptch.c - VGM Patcher
//

#include <stdio.h>
#include <stdlib.h>
#include "stdbool.h"
#include <string.h>
#include <math.h>	// for log(x) in GetVolModVal

#ifdef WIN32
#include <conio.h>
//#include <windows.h>
#endif

#include "zlib.h"

#include "VGMPlay.h"
#include "chip_strp.h"

// GD3 Tag Entry Count
#define GD3T_ENT_V100	11

int main(int argc, char* argv[]);
static signed char stricmp_u(const char *string1, const char *string2);
static signed char strnicmp_u(const char *string1, const char *string2, size_t count);
static unsigned long int GetFileLength(const char* FileName);
static bool OpenVGMFile(const char* FileName, bool* Compressed);
static bool WriteVGMFile(const char* FileName, bool Compress);
static unsigned char PreparseCommands(int ArgCount, char* ArgList[]);
static unsigned char ParseStripCommand(const char* StripCmd);
static unsigned char PatchVGM(int ArgCount, char* ArgList[]);
static unsigned char GetLoopModVal(char* CmdData);
static unsigned char GetVolModVal(char* CmdData);
static unsigned char CheckVGMFile(unsigned char Mode);
static bool ChipCommandIsUnknown(unsigned char Command);
static bool ChipCommandIsValid(unsigned char Command);
static unsigned long int RelocateVGMLoop(void);
static unsigned long int CalcGD3Length(unsigned long int DataPos, unsigned short int TagEntries);
static void StripVGMData(void);

VGM_HEADER VGMHead;
unsigned long int VGMDataLen;
unsigned char* VGMData;
STRIP_DATA StripVGM;

int main(int argc, char* argv[])
{
	int CmdCnt;
	int CurArg;
	// if SourceFile is NOT compressed, FileCompr = false -> DestFile will be uncompressed
	// if SourceFile IS compressed, FileCompr = true -> DestFile will also be compressed
	bool FileCompr;
	int ErrVal;
	unsigned char RetVal;
	
	printf("VGM Patcher\n-----------\n");
	
	ErrVal = 0;
	if (argc <= 0x01)
	{
		printf("Usage: vgm_ptch [-command1] [-command2] file1.vgm file2.vgz ...\n");
		printf("Use argument -help for command list.\n");
		goto EndProgram;
	}
	else if (! stricmp_u(argv[1], "-Help"))
	{
		printf("Help\n----\n");
		printf("Usage: vgm_ptch [-command1] [-command2] file1.vgm file2.vgz\n");
		printf("\n");
		printf("General Commands:\n");
		//printf("(no command)  like -ShowTag\n");
		printf("    -Help         Show this help\n");
		printf("    -ChipCmdList  List all supported chips with SetHzxx command\n");
		//printf("    -xxxxxxxxx  xxxxxxxxxxxxxxxxxx (following commands are ignored)\n");
		printf("\n");
		printf("Patching Commands:\n");
		printf("Command format: -command:value\n");
		printf("    -SetVer       Set Header-Version (e.g. 1.51)\n");
		printf("    -UpdateVer    Update Header-Version (additional changes if neccessary)\n");
		printf("    -SetRate      Sets the Playback rate (v1.01+)\n");
		printf("    -SetHzxxx     Sets the xxx's chip clock (see -ChipCmdList for details)\n");
		printf("    -SetLoopMod   Set the Loop Modifier (Format: *2, /2.0, 0x20)\n");
		printf("    -SetLoopBase  Set the Loop Base (-128 to 127)\n");
		printf("    -SetVolMod    Set the Volume Modifier (Format: 1.0, 0x00)\n");
		printf("\n");
#ifdef WIN32
		_getch();
#else
		getchar();
#endif
		printf("Commands to check the lengths (total, loop) and offsets (EOF, loop, GD3):\n");
		printf("    -Check        asks for correction\n");
		printf("    -CheckR       read only mode\n");
		printf("    -CheckL       autofix mode, recalculates lengths\n");
		printf("    -CheckO       like above, but tries to relocate loop offset\n");
		printf("\n");
		printf("    -Strip        Strip the data of a chip and/or channel\n");
		printf("                   Format: -Strip:Chip[:Ch,Ch,...];Chip\n");
		printf("                   e.g.: -Strip:PSG:0,1,2,Noise,Stereo;YM2151\n");
		printf("                   Note: Channel-Stripping isn't yet done.\n");
		printf("\n");
		printf("                   Tip: Stripping without commands optimizes the delays.\n");
		printf("\n");
		
		printf("Command names are case insensitive.\n");
		goto EndProgram;
	}
	else if (! stricmp_u(argv[1], "-ChipCmdList"))
	{
		printf("Chip Commands\n-------------\n");
		printf("    -SetHzPSG     Sets the SN76496 (PSG) chip clock\n");
		printf("                   Set to (Hz or 0xC0000000) for T6W28 (NGP) mode (v1.51+)\n");
		printf("    -SetHzYM2413  Sets the YM2413 (OPLL) chip clock\n");
		printf("VGM v1.10+:\n");
		printf("    -SetPSG_FdB   Sets the SN76496 Feedback Mask\n");
		printf("    -SetPSG_SRW   Sets the SN76496 Shift Register Width\n");
		printf("    -SetHzYM2612  Sets the YM2612 (OPN2) chip clock\n");
		printf("    -SetHzYM2151  Sets the YM2151 (OPM) chip clock\n");
		printf("VGM v1.51+:\n");
		printf("    -SetPSG_Flags Sets the SN76496 Flags\n");
		printf("    -SetHzSegaPCM Sets the SegaPCM chip clock\n");
		printf("    -SetSPCMIntf  Sets the SegaPCM interface register\n");
		printf("    -SetHzRF5C68  Sets the RF5C68 chip clock\n");
		printf("    -SetHzYM2203  Sets the YM2203 (OPN) chip clock\n");
		printf("    -SetHzYM2608  Sets the YM2608 (OPNA) chip clock\n");
		printf("    -SetHzYM2610  Sets the YM2610 (OPNB) chip clock\n");
		printf("                   Set to (Hz or 0x80000000) for YM2610B-mode\n");
		printf("    -SetHzYM3812  Sets the YM3812 (OPL2) chip clock\n");
		printf("    -SetHzYM3526  Sets the YM3526 (OPL) chip clock\n");
		printf("    -SetHzY8950   Sets the Y8950 (MSX AUDIO) chip clock\n");
#ifdef WIN32
		_getch();
#else
		getchar();
#endif
		printf("    -SetHzYMF262  Sets the YMF262 (OP3) chip clock\n");
		printf("    -SetHzYMF278B Sets the YMF278B (OPL4) chip clock\n");
		printf("    -SetHzYMF271  Sets the YMF271 (OPLX) chip clock\n");
		printf("    -SetHzYMZ280B Sets the YMZ280B chip clock\n");
		printf("    -SetHzRF5C164 Sets the RF5C164 (Sega MegaCD PCM) chip clock\n");
		printf("    -SetHzPWM     Sets the PWM chip clock\n");
		printf("    -SetHzAY8910  Sets the AY8910 chip clock\n");
		printf("\n");
		printf("Setting a clock rate to 0 disables the chip.\n");
		goto EndProgram;
	}
	
	for (CmdCnt = 0x01; CmdCnt < argc; CmdCnt ++)
	{
		if (*argv[CmdCnt] != '-')
			break;	// skip all commands
	}
	if (CmdCnt < 0x02)
	{
		printf("Error: No commands specified!\n");
		goto EndProgram;
	}
	if (CmdCnt >= argc)
	{
		printf("Error: No files specified!\n");
		goto EndProgram;
	}
	
	PreparseCommands(CmdCnt - 0x01, argv + 0x01);
	for (CurArg = CmdCnt; CurArg < argc; CurArg ++)
	{
		printf("File: %s ...\n", argv[CurArg]);
		if (! OpenVGMFile(argv[CurArg], &FileCompr))
		{
			printf("Error opening file %s!\n", argv[CurArg]);
			printf("\n");
			ErrVal |= 1;	// There was at least 1 opening-error.
			continue;
		}
		
		RetVal = PatchVGM(CmdCnt - 0x01, argv + 0x01);
		if (RetVal == 0x10)
		{
			// Showed Tag - no write neccessary
		}
		else if (RetVal)
		{
			if (RetVal == 0x80)
			{
				ErrVal |= 8;
				goto EndProgram;	// Argument Error
			}
			ErrVal |= 4;	// At least 1 file wasn't patched.
		}
		else //if (! RetVal)
		{
			if (! WriteVGMFile(argv[CurArg], FileCompr))
			{
				printf("Error opening file %s!\n", argv[CurArg]);
				ErrVal |= 2;	// There was at least 1 writing-error.
			}
		}
		printf("\n");
	}
	
EndProgram:
#ifdef WIN32
	if (argv[0][1] == ':')
	{
		// Executed by Double-Clicking (or Drap and Drop)
		if (_kbhit())
			_getch();
		_getch();
	}
#endif
	
	return ErrVal;
}

static signed char stricmp_u(const char *string1, const char *string2)
{
	// my own stricmp, because VC++6 doesn't find _stricmp when compiling without
	// standard libraries
	const char* StrPnt1;
	const char* StrPnt2;
	char StrChr1;
	char StrChr2;
	
	StrPnt1 = string1;
	StrPnt2 = string2;
	while(true)
	{
		StrChr1 = toupper(*StrPnt1);
		StrChr2 = toupper(*StrPnt2);
		
		if (StrChr1 < StrChr2)
			return -1;
		else if (StrChr1 > StrChr2)
			return +1;
		if (StrChr1 == 0x00)
			return 0;
		
		StrPnt1 ++;
		StrPnt2 ++;
	}
	
	return 0;
}

static signed char strnicmp_u(const char *string1, const char *string2, size_t count)
{
	// my own strnicmp, because GCC doesn't seem to have _strnicmp
	const char* StrPnt1;
	const char* StrPnt2;
	char StrChr1;
	char StrChr2;
	size_t CurChr;
	
	StrPnt1 = string1;
	StrPnt2 = string2;
	CurChr = 0x00;
	while(CurChr < count)
	{
		StrChr1 = toupper(*StrPnt1);
		StrChr2 = toupper(*StrPnt2);
		
		if (StrChr1 < StrChr2)
			return -1;
		else if (StrChr1 > StrChr2)
			return +1;
		if (StrChr1 == 0x00)
			return 0;
		
		StrPnt1 ++;
		StrPnt2 ++;
		CurChr ++;
	}
	
	return 0;
}

static unsigned long int GetFileLength(const char* FileName)
{
	FILE* hFile;
	unsigned long int FileSize;
	
	hFile = fopen(FileName, "rb");
	if (hFile == NULL)
		return 0xFFFFFFFF;
	
	fseek(hFile, 0x00, SEEK_END);
	FileSize = ftell(hFile);
	
	fclose(hFile);
	
	return FileSize;
}

static bool OpenVGMFile(const char* FileName, bool* Compressed)
{
	gzFile hFile;
	unsigned long int FileSize;
	unsigned long int fccHeader;
	unsigned long int CurPos;
	
	FileSize = GetFileLength(FileName);
	
	hFile = gzopen(FileName, "rb");
	if (hFile == NULL)
		return false;
	
	gzseek(hFile, 0x00, SEEK_SET);
	gzread(hFile, &fccHeader, 0x04);
	if (fccHeader != FCC_VGM)
		goto OpenErr;
	
	*Compressed = ! gzdirect(hFile);
	
	gzseek(hFile, 0x00, SEEK_SET);
	gzread(hFile, &VGMHead, sizeof(VGM_HEADER));
	
	// I'll skip the Header preperations. I'll deal with that later
	// Header preperations
	/*if (VGMHead.lngVersion < 0x00000101)
	{
		VGMHead.lngRate = 0;
	}
	if (VGMHead.lngVersion < 0x00000110)
	{
		VGMHead.shtPSG_Feedback = 0x0000;
		VGMHead.bytPSG_SRWidth = 0x00;
		VGMHead.lngHzYM2612 = VGMHead.lngHzYM2413;
		VGMHead.lngHzYM2151 = VGMHead.lngHzYM2413;
	}
	if (VGMHead.lngVersion < 0x00000150)
	{
		VGMHead.lngDataOffset = 0x00000000;
	}
	if (VGMHead.lngVersion < 0x00000151)
	{
		VGMHead.lngHzSPCM = 0x0000;
		VGMHead.lngSPCMIntf = 0x00000000;
		// all others are zeroed by memset
	}*/
	// relative -> absolute addresses
	/*VGMHead.lngEOFOffset += 0x00000004;
	if (VGMHead.lngGD3Offset)
		VGMHead.lngGD3Offset += 0x00000014;
	if (VGMHead.lngLoopOffset)
		VGMHead.lngLoopOffset += 0x0000001C;
	if (! VGMHead.lngDataOffset)
		VGMHead.lngDataOffset = 0x0000000C;
	VGMHead.lngDataOffset += 0x00000034;*/
	
	CurPos = VGMHead.lngDataOffset;
	if (VGMHead.lngVersion < 0x00000150)
		CurPos = 0x40;
	else
		CurPos = 0x00000034 + VGMHead.lngDataOffset;
	memset((unsigned char*)&VGMHead + CurPos, 0x00, sizeof(VGM_HEADER) - CurPos);
	
	// Read Data
	if (*Compressed)
		VGMDataLen = 0x00000004 + VGMHead.lngEOFOffset;	// size from EOF offset
	else
		VGMDataLen = FileSize;	// size of the actual file
	VGMData = (unsigned char*)malloc(VGMDataLen);
	if (VGMData == NULL)
		goto OpenErr;
	gzseek(hFile, 0x00, SEEK_SET);
	gzread(hFile, VGMData, VGMDataLen);
	
	gzclose(hFile);
	
	return true;

OpenErr:

	gzclose(hFile);
	return false;
}

static bool WriteVGMFile(const char* FileName, bool Compress)
{
	gzFile hFile;
	
	if (! Compress)
		hFile = fopen(FileName, "wb");
	else
		hFile = gzopen(FileName, "wb9");
	if (hFile == NULL)
		return false;
	
	// Write VGM Data (including GD3 Tag)
	if (! Compress)
	{
		fseek((FILE*)hFile, 0x00, SEEK_SET);
		fwrite(VGMData, 0x01, VGMDataLen, (FILE*)hFile);
	}
	else
	{
		gzseek(hFile, 0x00, SEEK_SET);
		gzwrite(hFile, VGMData, VGMDataLen);
	}
	
	if (! Compress)
		fclose((FILE*)hFile);
	else
		gzclose(hFile);
	printf("File written.\n");
	
	return true;
}

static unsigned char PreparseCommands(int ArgCount, char* ArgList[])
{
	int CurArg;
	char* CmdStr;
	char* CmdData;
	unsigned char RetVal;
	char ChrBak;
	//unsigned char TempByt;
	//unsigned short int TempSht;
	//unsigned long int TempLng;
	
	memset(&StripVGM, 0x00, sizeof(STRIP_DATA));
	for (CurArg = 0x00; CurArg < ArgCount; CurArg ++)
	{
		CmdStr = ArgList[CurArg] + 0x01;	// Skip the '-' at the beginning
		
		CmdData = strchr(CmdStr, ':');
		if (CmdData != NULL)
		{
			// actually this is quite dirty ...
			ChrBak = *CmdData;
			*CmdData = 0x00;
			CmdData ++;
		}
		
		/*if (CmdData != NULL)
			printf("%s: %s\n", CmdStr, CmdData);
		else
			printf("%s\n", CmdStr);*/
		if (! stricmp_u(CmdStr, "Strip"))
		{
			RetVal = ParseStripCommand(CmdData);
			if (RetVal)
				return RetVal;
			// TODO: Parse "Strip"-Command
			//StripVGM.SN76496.All = true;
			//StripVGM.YM2612.All = true;
		}
		if (CmdData != NULL)
		{
			// ... and this even more
			CmdData --;
			*CmdData = ChrBak;
		}
	}
	
	return 0x00;
}

static unsigned char ParseStripCommand(const char* StripCmd)
{
	char* StripTmp;
	char* ChipPos;
	char* ChnPos;
	char* NxtChip;
	//char* TempStr;
	unsigned char StripMode;
	unsigned char CurChip;
	STRIP_GENERIC* TempChip;
	//unsigned char TempByt;
	//unsigned short int TempSht;
	//unsigned long int TempLng;
	
	if (StripCmd == NULL)
		return 0x00;	// Passing no argument is valid
	
	// Format: "ChipA:Chn1,Chn2,Chn3;ChipB"
	StripTmp = (char*)malloc(strlen(StripCmd) + 0x01);
	strcpy(StripTmp, StripCmd);
	ChipPos = StripTmp;
	while(ChipPos != NULL)
	{
		ChnPos = strchr(ChipPos, ':');
		NxtChip = strchr(ChipPos, ';');
		StripMode = 0x00;
		if (NxtChip != NULL && ChnPos != NULL)
		{
			if (ChnPos < NxtChip)
				StripMode = 0x01;	// Strip Channel x,x,x
			else
				StripMode = 0x00;	// Strip All
		}
		else if (ChnPos == NULL)	// && (NxtChip != NULL || NxtChip == NULL)
		{
			StripMode = 0x00;	// Strip All
		}
		else if (ChnPos != NULL)	// && NxtChip == NULL
		{
			StripMode = 0x01;	// Last Entry - Strip Channels
		}
		
		switch(StripMode)
		{
		case 0x00:	// 00 - Strip All
			if (NxtChip != NULL)
			{
				*NxtChip = 0x00;
				NxtChip ++;
			}
			break;
		case 0x01:	// 01 - Strip Channels
			*ChnPos = 0x00;	// ChnPos can't be NULL
			ChnPos ++;
			break;
		}
		
		// I can compare ChipPos, because I zeroed the split-char
		if (! stricmp_u(ChipPos, "PSG") || ! stricmp_u(ChipPos, "SN76496"))
		{
			CurChip = 0x00;
			TempChip = (STRIP_GENERIC*)&StripVGM.SN76496;
		}
		else if (! stricmp_u(ChipPos, "YM2413"))
		{
			CurChip = 0x01;
			TempChip = (STRIP_GENERIC*)&StripVGM.YM2413;
		}
		else if (! stricmp_u(ChipPos, "YM2612"))
		{
			CurChip = 0x02;
			TempChip = (STRIP_GENERIC*)&StripVGM.YM2612;
		}
		else if (! stricmp_u(ChipPos, "YM2151"))
		{
			CurChip = 0x03;
			TempChip = (STRIP_GENERIC*)&StripVGM.YM2151;
		}
		else if (! stricmp_u(ChipPos, "SegaPCM"))
		{
			CurChip = 0x04;
			TempChip = (STRIP_GENERIC*)&StripVGM.SegaPCM;
		}
		else if (! stricmp_u(ChipPos, "RF5C68"))
		{
			CurChip = 0x05;
			TempChip = (STRIP_GENERIC*)&StripVGM.RF5C68;
		}
		else if (! stricmp_u(ChipPos, "YM2203"))
		{
			CurChip = 0x06;
			TempChip = (STRIP_GENERIC*)&StripVGM.YM2203;
		}
		else if (! stricmp_u(ChipPos, "YM2608"))
		{
			CurChip = 0x07;
			TempChip = (STRIP_GENERIC*)&StripVGM.YM2608;
		}
		else if (! stricmp_u(ChipPos, "YM2610"))
		{
			CurChip = 0x08;
			TempChip = (STRIP_GENERIC*)&StripVGM.YM2610;
		}
		else if (! stricmp_u(ChipPos, "YM3812"))
		{
			CurChip = 0x09;
			TempChip = (STRIP_GENERIC*)&StripVGM.YM3812;
		}
		else if (! stricmp_u(ChipPos, "YM3526"))
		{
			CurChip = 0x0A;
			TempChip = (STRIP_GENERIC*)&StripVGM.YM3526;
		}
		else if (! stricmp_u(ChipPos, "Y8950"))
		{
			CurChip = 0x0B;
			TempChip = (STRIP_GENERIC*)&StripVGM.Y8950;
		}
		else if (! stricmp_u(ChipPos, "YMF262"))
		{
			CurChip = 0x0C;
			TempChip = (STRIP_GENERIC*)&StripVGM.YMF262;
		}
		else if (! stricmp_u(ChipPos, "YMF278B"))
		{
			CurChip = 0x0D;
			TempChip = (STRIP_GENERIC*)&StripVGM.YMF278B_FM;
		}
		else if (! stricmp_u(ChipPos, "YMF271"))
		{
			CurChip = 0x0E;
			TempChip = (STRIP_GENERIC*)&StripVGM.YMF271;
		}
		else if (! stricmp_u(ChipPos, "YMZ280B"))
		{
			CurChip = 0x0F;
			TempChip = (STRIP_GENERIC*)&StripVGM.YMZ280B;
		}
		else if (! stricmp_u(ChipPos, "RF5C164"))
		{
			CurChip = 0x10;
			TempChip = (STRIP_GENERIC*)&StripVGM.RF5C164;
		}
		else if (! stricmp_u(ChipPos, "PWM"))
		{
			CurChip = 0x11;
			TempChip = (STRIP_GENERIC*)&StripVGM.PWM;
		}
		else if (! stricmp_u(ChipPos, "AY8910"))
		{
			CurChip = 0x12;
			TempChip = (STRIP_GENERIC*)&StripVGM.AY8910;
		}
		else
		{
			printf("Strip Error - Unknown Chip: %s\n", ChipPos);
			return 0x80;
		}
		
		switch(StripMode)
		{
		case 0x00:	// 00 - Strip All
			if (CurChip != 0x0D)	// the OPL4 is special
				TempChip->All = true;
			else
				StripVGM.YMF278B_All = true;
			break;
		case 0x01:	// 01 - Strip Channels
			ChnPos ++;
			// TempStr contains the Channel-List
			printf("TODO ...\n");
			return 0x10;
			break;
		}
		
		ChipPos = NxtChip;
	}
	
	return 0x00;
}

static unsigned char PatchVGM(int ArgCount, char* ArgList[])
{
	int CurArg;
	char* CmdStr;
	char* CmdData;
	//unsigned long int DataLen;
	char ChrBak;
	char* TempPnt;
	char TempStr[0x03];
	unsigned char TempByt;
	unsigned short int TempSht;
	unsigned long int TempLng;
	signed long int TempSLng;
	unsigned char RetVal;
	unsigned long int OldVal;
	unsigned long int NewVal;
	unsigned long int* ChipHzPnt;
	
	// Execute Commands
	RetVal = 0x10;	// nothing done - skip writing
	//if (! ArgCount)
	//	ShowHeader();
	for (CurArg = 0x00; CurArg < ArgCount; CurArg ++)
	{
		CmdStr = ArgList[CurArg] + 0x01;	// Skip the '-' at the beginning
		
		CmdData = strchr(CmdStr, ':');
		if (CmdData != NULL)
		{
			// and the same dirt here again
			ChrBak = *CmdData;
			*CmdData = 0x00;
			CmdData ++;
		}
		
		if (CmdData != NULL)
			printf("%s: %s\n", CmdStr, CmdData);
		else
			printf("%s\n", CmdStr);
		if (! stricmp_u(CmdStr, "SetVer") || ! stricmp_u(CmdStr, "UpdateVer"))
		{
			if (CmdData == NULL)
				goto IncompleteArg;
			
			OldVal = VGMHead.lngVersion;
			
			TempPnt  = strchr(CmdData, '.');
			TempLng = strtoul(CmdData, NULL, 0x10);	// Version Major
			if (TempPnt)
			{
				strncpy(TempStr, TempPnt + 0x01, 0x02);
				TempStr[0x02] = 0x00;
				if (! TempStr[0x01])
					TempStr[0x01] = '0';
				TempByt = (unsigned char)strtoul(TempStr, NULL, 0x10);	// Version Minor
			}
			else
			{
				TempByt = 0x00;
			}
			NewVal = (TempLng << 8) | (TempByt << 0);
			
			if (OldVal != NewVal)
			{
				VGMHead.lngVersion = NewVal;
				RetVal = 0x00;
				
				if (! stricmp_u(CmdStr, "UpdateVer"))
				{
					// Up- and downdate to new version
					if (OldVal < NewVal)
					{
						// Update version (write missing header data)
						if (OldVal < 0x00000101 && NewVal >= 0x00000101)	// v1.00 -> v1.01+
						{
							VGMHead.lngRate = 0;
						}
						if (OldVal < 0x00000110 && NewVal >= 0x00000110)	// v1.01 -> v1.10+
						{
							VGMHead.shtPSG_Feedback = 0x0009;
							VGMHead.bytPSG_SRWidth = 0x10;
							VGMHead.lngHzYM2612 = VGMHead.lngHzYM2413;
							VGMHead.lngHzYM2151 = VGMHead.lngHzYM2413;
							// TODO: Scan vgm for used chips
						}
						if (OldVal < 0x00000150 && NewVal >= 0x00000150)	// v1.10 -> v1.50+
						{
							VGMHead.lngDataOffset = 0x0000000C;
						}
						if (OldVal < 0x00000151 && NewVal >= 0x00000151)	// v1.50 -> v1.51+
						{
							VGMHead.bytPSG_Flags = (0x00 << 0) | (0x00 << 1) | (0x00 << 2);
							VGMHead.lngHzSPCM = 0x0000;
							VGMHead.lngSPCMIntf = 0x00000000;
							// all others are zeroed by memset during OpenVGMFile
						}
					}
					else if (OldVal > NewVal)
					{
						// Downdate version (strip unneeded header data)
						if (OldVal >= 0x00000151 && NewVal < 0x00000151)	// v1.51+ -> v1.50
						{
							VGMHead.bytPSG_Flags = 0x00;
							TempLng = sizeof(VGM_HEADER) - 0x38;
							memset(&VGMHead.lngHzSPCM, 0x00, TempLng);
						}
						if (OldVal >= 0x00000150 && NewVal < 0x00000150)	// v1.50+ -> v1.10
						{
							if (VGMHead.lngDataOffset != 0x0000000C)
							{
								// I'm just too lazy to move all data
								// and no one would even THINK of downdating vgms
								// Update: Except for me ...
								//         (I wanted v1.51 -> 1.10 -> 1.50 for a smaller header)
								TempLng = VGMHead.lngDataOffset + 0x00000034;
								TempSLng = VGMHead.lngDataOffset - 0x0000000C;
								VGMHead.lngEOFOffset -= TempSLng;
								if (VGMHead.lngGD3Offset)
									VGMHead.lngGD3Offset -= TempSLng;
								if (VGMHead.lngLoopOffset)
									VGMHead.lngLoopOffset -= TempSLng;
								
								memmove(&VGMData[0x40], &VGMData[TempLng], VGMDataLen - TempLng);
								VGMDataLen -= TempSLng;
							}
							VGMHead.lngDataOffset = 0x0000000;
						}
						if (OldVal >= 0x00000110 && NewVal < 0x00000110)	// v1.10+ -> v1.01
						{
							VGMHead.shtPSG_Feedback = 0x0000;
							VGMHead.bytPSG_SRWidth = 0x00;
							if (! VGMHead.lngHzYM2413)
							{
								if (VGMHead.lngHzYM2151)
									VGMHead.lngHzYM2413 = VGMHead.lngHzYM2151;
								else if (VGMHead.lngHzYM2612)
									VGMHead.lngHzYM2413 = VGMHead.lngHzYM2612;
							}
						}
						if (OldVal >= 0x00000101 && NewVal < 0x00000101)	// v1.01+ -> v1.00
						{
							VGMHead.lngRate = 0;
						}
					}
				}
			}
		}
		else if (! stricmp_u(CmdStr, "SetRate"))
		{
			if (CmdData == NULL)
				goto IncompleteArg;
			
			NewVal = strtoul(CmdData, NULL, 0);
			if (NewVal != VGMHead.lngRate)
			{
				VGMHead.lngRate = NewVal;
				RetVal = 0x00;
			}
		}
		else if (! strnicmp_u(CmdStr, "SetHz", 5))
		{
			if (CmdData == NULL)
				goto IncompleteArg;
			
			ChipHzPnt = NULL;
			if (! stricmp_u(CmdStr + 5, "PSG"))
			{
				ChipHzPnt = &VGMHead.lngHzPSG;
				OldVal = 0x00000100;
			}
			else if (! stricmp_u(CmdStr + 5, "YM2413"))
			{
				ChipHzPnt = &VGMHead.lngHzYM2413;
				OldVal = 0x00000100;
			}
			else if (! stricmp_u(CmdStr + 5, "YM2612"))
			{
				ChipHzPnt = &VGMHead.lngHzYM2612;
				OldVal = 0x00000110;
			}
			else if (! stricmp_u(CmdStr + 5, "YM2151"))
			{
				ChipHzPnt = &VGMHead.lngHzYM2151;
				OldVal = 0x00000110;
			}
			else if (! stricmp_u(CmdStr + 5, "SegaPCM"))
			{
				ChipHzPnt = &VGMHead.lngHzSPCM;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "RF5C68"))
			{
				ChipHzPnt = &VGMHead.lngHzRF5C68;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "YM2203"))
			{
				ChipHzPnt = &VGMHead.lngHzYM2203;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "YM2608"))
			{
				ChipHzPnt = &VGMHead.lngHzYM2608;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "YM2610"))
			{
				ChipHzPnt = &VGMHead.lngHzYM2610;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "YM3812"))
			{
				ChipHzPnt = &VGMHead.lngHzYM3812;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "YM3526"))
			{
				ChipHzPnt = &VGMHead.lngHzYM3526;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "Y8950"))
			{
				ChipHzPnt = &VGMHead.lngHzY8950;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "YMF262"))
			{
				ChipHzPnt = &VGMHead.lngHzYMF262;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "YMF278B"))
			{
				ChipHzPnt = &VGMHead.lngHzYMF278B;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "YMF271"))
			{
				ChipHzPnt = &VGMHead.lngHzYMF271;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "YMZ280B"))
			{
				ChipHzPnt = &VGMHead.lngHzYMZ280B;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "RF5C164"))
			{
				ChipHzPnt = &VGMHead.lngHzRF5C164;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "PWM"))
			{
				ChipHzPnt = &VGMHead.lngHzPWM;
				OldVal = 0x00000151;
			}
			else if (! stricmp_u(CmdStr + 5, "AY8910"))
			{
				ChipHzPnt = &VGMHead.lngHzAY8910;
				OldVal = 0x00000151;
			}
			else
			{
				printf("Error - Unknown Chip: %s\n", CmdStr + 5);
				return 0x80;
			}
			
			if (VGMHead.lngVersion >= OldVal)
			{
				NewVal = strtoul(CmdData, NULL, 0);
				if (NewVal != *ChipHzPnt)
				{
					*ChipHzPnt = NewVal;
					RetVal = 0x00;
				}
			}
			else
			{
				printf("Warning! Command ignored - VGM version too low!\n");
			}
		}
		else if (! strnicmp_u(CmdStr, "SetPSG_", 7))
		{
			if (CmdData == NULL)
				goto IncompleteArg;
			
			if (! stricmp_u(CmdStr + 7, "FdB"))
			{
				if (VGMHead.lngVersion >= 0x00000110)
				{
					TempSht = (unsigned short int)strtoul(CmdData, NULL, 0);
					if (TempSht != VGMHead.shtPSG_Feedback)
					{
						VGMHead.shtPSG_Feedback = TempSht;
						RetVal = 0x00;
					}
				}
				else
				{
					printf("Warning! Command ignored - VGM version too low!\n");
				}
			}
			else if (! stricmp_u(CmdStr + 7, "SRW"))
			{
				if (VGMHead.lngVersion >= 0x00000110)
				{
					TempByt = (unsigned char)strtoul(CmdData, NULL, 0);
					if (TempByt != VGMHead.bytPSG_SRWidth)
					{
						VGMHead.bytPSG_SRWidth = TempByt;
						RetVal = 0x00;
					}
				}
				else
				{
					printf("Warning! Command ignored - VGM version too low!\n");
				}
			}
			else if (! stricmp_u(CmdStr + 7, "Flag"))
			{
				if (VGMHead.lngVersion >= 0x00000151)
				{
					TempByt = (unsigned char)strtoul(CmdData, NULL, 0);
					if (TempByt != VGMHead.bytPSG_Flags)
					{
						VGMHead.bytPSG_Flags = TempByt;
						RetVal = 0x00;
					}
				}
				else
				{
					printf("Warning! Command ignored - VGM version too low!\n");
				}
			}
			else
			{
				printf("Error - Unknown Command: -%s\n", CmdStr);
				return 0x80;
			}
		}
		else if (! stricmp_u(CmdStr, "SetSPCMIntf"))
		{
			if (CmdData == NULL)
				goto IncompleteArg;
			
			if (VGMHead.lngVersion >= 0x00000151)
			{
				NewVal = strtoul(CmdData, NULL, 0);
				if (NewVal != VGMHead.lngSPCMIntf)
				{
					VGMHead.lngSPCMIntf = NewVal;
					RetVal = 0x00;
				}
			}
			else
			{
				printf("Warning! Command ignored - VGM version too low!\n");
			}
		}
		else if (! stricmp_u(CmdStr, "SetLoopMod"))
		{
			if (CmdData == NULL)
				goto IncompleteArg;
			
			if (VGMHead.lngVersion >= 0x00000151)
			{
				TempByt = GetLoopModVal(CmdData);
				if (TempByt != VGMHead.bytLoopModifier)
				{
					VGMHead.bytLoopModifier = (unsigned char)TempByt;
					RetVal = 0x00;
				}
			}
			else
			{
				printf("Warning! Command ignored - VGM version too low!\n");
			}
		}
		else if (! stricmp_u(CmdStr, "SetLoopBase"))
		{
			if (CmdData == NULL)
				goto IncompleteArg;
			
			if (VGMHead.lngVersion >= 0x00000151)
			{
				TempSLng = strtol(CmdData, NULL, 0);
				if (TempSLng != VGMHead.bytLoopBase)
				{
					VGMHead.bytLoopBase = (signed char)TempSLng;
					RetVal = 0x00;
				}
			}
			else
			{
				printf("Warning! Command ignored - VGM version too low!\n");
			}
		}
		else if (! stricmp_u(CmdStr, "SetVolMod"))
		{
			if (CmdData == NULL)
				goto IncompleteArg;
			
			if (VGMHead.lngVersion >= 0x00000151 || VGMHead.lngDataOffset >= 0x00000049)
			{
				TempByt = GetVolModVal(CmdData);
				if (TempByt != VGMHead.bytVolumeModifier)
				{
					VGMHead.bytVolumeModifier = TempByt;
					RetVal = 0x00;
				}
			}
			else
			{
				printf("Warning! Command ignored - VGM version too low!\n");
			}
		}
		else if (! strnicmp_u(CmdStr, "Check", 5))
		{
			if (! stricmp_u(CmdStr + 5, ""))
			{
				// ask for way to correct
				TempByt = 0x03;
			}
			else if (! stricmp_u(CmdStr + 5, "R"))
			{
				// read only
				TempByt = 0x00;
			}
			else if (! stricmp_u(CmdStr + 5, "L"))
			{
				// fix by recalculating length
				TempByt = 0x01;
			}
			else if (! stricmp_u(CmdStr + 5, "O"))
			{
				// fix by relocating loop offset
				TempByt = 0x02;
			}
			else
			{
				printf("Error - Unknown Command: -%s\n", CmdStr);
				return 0x80;
			}
			
			TempByt = CheckVGMFile(TempByt);
			if (TempByt == 0xFF)
			{
				printf("Error processing vgm file!\n");
				return 0xFF;
			}
			else if (TempByt)
			{
				RetVal = 0x00;
			}
		}
		else if (! stricmp_u(CmdStr, "Strip"))
		{
			printf("Stripping data ...");
			StripVGMData();
			RetVal = 0x00;
		}
		else
		{
			printf("Error - Unknown Command: -%s\n", CmdStr);
			return 0x80;
		}
		
		if (CmdData != NULL)
		{
			CmdData --;
			*CmdData = ChrBak;
		}
	}
	
	if (RetVal != 0x10)
	{
		// Write VGM Header
		if (VGMHead.lngVersion < 0x00000150)
			TempLng = 0x40;
		else
			TempLng = 0x00000034 + VGMHead.lngDataOffset;
		memcpy(&VGMData[0x00], &VGMHead, TempLng);
	}
	
	return RetVal;

IncompleteArg:

	printf("Error! Argument incomplete!\n");
	
	return 0x80;
}

static unsigned char GetLoopModVal(char* CmdData)
{
	double LoopDbl;
	signed long int LoopLng;
	
	switch(CmdData[0x00])
	{
	case '*':
	case '/':
		// *2, /2.0
		LoopDbl = strtod(CmdData + 0x01,  NULL);
		switch(CmdData[0x00])
		{
		case '*':
			LoopLng = (unsigned long int)floor(LoopDbl * 16.0 + 0.5);
			break;
		case '/':
			LoopLng = (unsigned long int)floor(16.0 / LoopDbl + 0.5);
			break;
		}
		if (LoopLng == 0x00)
			LoopLng = 0x01;
		break;
	default:
		// 0x00 is accepted here, but only here
		LoopLng = strtol(CmdData, NULL, 0);
		break;
	}
	
	// Note 0x00 is valid, but equal to 0x10, so minimum is 0x01
	if (LoopLng < 0x00)
		LoopLng = 0x01;
	else if (LoopLng > 0xFF)
		LoopLng = 0xFF;
	
	return (unsigned char)LoopLng;
}

static unsigned char GetVolModVal(char* CmdData)
{
	char* EndStr;
	double VolDbl;
	signed long int VolLng;
	
	VolLng = strtol(CmdData, &EndStr, 0);
	if (*EndStr == '.')
	{
		// 1.0
		VolDbl = strtod(CmdData, NULL);
		if (VolDbl < 0.25)
			VolDbl = 0.25;
		else if (VolDbl > 64.0)
			VolDbl = 64.0;
		VolLng = (signed long int)floor(log(VolDbl) / log(2.0) * 0x20 + 0.5);
	}
	
	if (VolLng < 0x00)
	{
		if (VolLng < -0x3F)
			VolLng = -0x3F;
		VolLng = 0x100 + VolLng;
	}
	else if (VolLng > 0xFF)
	{
		VolLng = 0x00;
	}
	
	return (unsigned char)VolLng;
}

static unsigned char CheckVGMFile(unsigned char Mode)
{
	unsigned long int CurPos;
	unsigned char Command;
	unsigned char TempByt;
	//unsigned short int TempSht;
	unsigned long int TempLng;
	unsigned long int CmdLen;
	bool StopVGM;
	bool HasLoop;
	bool InLoop;
	unsigned long int CountLen;
	unsigned long int CountLoop;
	unsigned long int VGMLoop;
	unsigned long int LoopPos;
	unsigned short int CmdDelay;
	unsigned long int VGMEoDPos;	// EoD - End of Data
	unsigned long int VGMGD3Pos;
	unsigned long int GD3Ver;
	unsigned char VGMErr;
	unsigned char GD3Flags;
	unsigned char RetVal;
	bool BadCmdFound;
	unsigned long int NewDataPos;
	
	HasLoop = VGMHead.lngLoopOffset ? true : false;
	if (! HasLoop && VGMHead.lngLoopSamples)
	{
		printf("LoopOffset missing, but LoopSamples > 0!\n");
		HasLoop = true;
		printf("Assume song to have a loop!\n");
	}
	if (HasLoop)
		VGMLoop = VGMHead.lngTotalSamples - VGMHead.lngLoopSamples;
	if (VGMHead.lngLoopSamples)
		LoopPos = 0x0000001C + VGMHead.lngLoopOffset;
	else
		LoopPos = 0x00000000;
	CountLen = 0x00;
	CountLoop = 0x00;
	
	VGMErr = 0x00;
	InLoop = false;
	BadCmdFound = false;
	if (VGMHead.lngVersion < 0x00000150)
	{
		CurPos = 0x40;
	}
	else
	{
		CurPos = 0x00000034 + VGMHead.lngDataOffset;
		Command = VGMData[CurPos + 0x00];
		if (! VGMHead.lngDataOffset || ! ChipCommandIsValid(Command))
		{
			VGMErr |= 0x04;
			printf("Bad Data Offset!\n");
			
			StopVGM = false;
			while(CurPos < VGMDataLen)
			{
				Command = VGMData[CurPos + 0x00];
				if (Command >= 0x70 && Command <= 0x8F)
				{
					StopVGM = true;
					CmdLen = 0x01;
				}
				else
				{
					switch(Command)
					{
					case 0x66:	// End Of File
						StopVGM = true;
						break;
					case 0x62:	// 1/60s delay
						StopVGM = true;
						break;
					case 0x63:	// 1/50s delay
						StopVGM = true;
						break;
					case 0x61:	// xx Sample Delay
						StopVGM = true;
						break;
					case 0x50:	// SN76496 write
						CmdLen = 0x02;
						StopVGM = ChipCommandIsValid(Command);
						break;
					case 0x67:	// Data Block (PCM Data Stream)
						TempByt = VGMData[CurPos + 0x01];
						memcpy(&TempLng, &VGMData[CurPos + 0x03], 0x04);
						CmdLen = 0x07 + TempLng;
						
						if (TempByt == 0x66 && CurPos + CmdLen <= VGMDataLen)
							StopVGM = true;
						else
							CmdLen = 0x01;	// invalid block
						break;
					case 0x68:	// PCM RAM write
						TempByt = VGMData[CurPos + 0x01];
						CmdLen = 0x0C;
						if (TempByt == 0x66)
							StopVGM = true;
						else
							CmdLen = 0x01;	// invalid block
						break;
					case 0x90:	// DAC Ctrl: Setup Chip
					case 0x91:	// DAC Ctrl: Set Data
					case 0x92:	// DAC Ctrl: Set Freq
					case 0x93:	// DAC Ctrl: Play from Start Pos
					case 0x94:	// DAC Ctrl: Stop immediately
					case 0x95:	// DAC Ctrl: Play Block (small)
						StopVGM = true;
						break;
					default:	// Handle all other commands
						switch(Command & 0xF0)
						{
						case 0x30:
						case 0x40:
							CmdLen = 0x02;
							StopVGM = ChipCommandIsValid(Command);
							break;
						case 0x50:
						case 0xA0:
						case 0xB0:
							CmdLen = 0x03;
							StopVGM = ChipCommandIsValid(Command);
							break;
						case 0xC0:
						case 0xD0:
							CmdLen = 0x04;
							StopVGM = ChipCommandIsValid(Command);
							break;
						case 0xE0:
						case 0xF0:
							CmdLen = 0x05;
							StopVGM = ChipCommandIsValid(Command);
							break;
						default:
							CmdLen = 0x01;
							break;
						}
						break;
					}
				}
				if (StopVGM)
					break;
				
				CurPos += CmdLen;
			}
			NewDataPos = CurPos;
		}
		
		CurPos = 0x00000034 + VGMHead.lngDataOffset;
	}
	
	StopVGM = false;
	while(CurPos < VGMDataLen)	// I have my reasons for NOT using VGMHead.lngEOFOffset
	{
		if (HasLoop)
		{
			if (CurPos == LoopPos)
				InLoop = true;
		}
		
		CmdLen = 0x00;
		CmdDelay = 0x00;
		Command = VGMData[CurPos + 0x00];
		if (Command >= 0x70 && Command <= 0x8F)
		{
			switch(Command & 0xF0)
			{
			case 0x70:
				CmdDelay = (Command & 0x0F) + 0x01;
				break;
			case 0x80:
				CmdDelay = Command & 0x0F;
				break;
			}
			CmdLen = 0x01;
		}
		else
		{
			switch(Command)
			{
			case 0x66:	// End Of File
				CmdLen = 0x01;
				StopVGM = true;
				break;
			case 0x62:	// 1/60s delay
				CmdDelay = 735;
				CmdLen = 0x01;
				break;
			case 0x63:	// 1/50s delay
				CmdDelay = 882;
				CmdLen = 0x01;
				break;
			case 0x61:	// xx Sample Delay
				memcpy(&CmdDelay, &VGMData[CurPos + 0x01], 0x02);
				CmdLen = 0x03;
				break;
			case 0x50:	// SN76496 write
				if (! ChipCommandIsValid(Command))
					printf("Warning! Command %02X found, but associated chip not used!", Command);
				CmdLen = 0x02;
				break;
			case 0x67:	// Data Block (PCM Data Stream)
				TempByt = VGMData[CurPos + 0x02];
				memcpy(&TempLng, &VGMData[CurPos + 0x03], 0x04);
				CmdLen = 0x07 + TempLng;
				break;
			case 0x68:	// PCM RAM write
				CmdLen = 0x0C;
				break;
			case 0x90:	// DAC Ctrl: Setup Chip
				CmdLen = 0x05;
				break;
			case 0x91:	// DAC Ctrl: Set Data
				CmdLen = 0x05;
				break;
			case 0x92:	// DAC Ctrl: Set Freq
				CmdLen = 0x06;
				break;
			case 0x93:	// DAC Ctrl: Play from Start Pos
				CmdLen = 0x0B;
				break;
			case 0x94:	// DAC Ctrl: Stop immediately
				CmdLen = 0x02;
				break;
			case 0x95:	// DAC Ctrl: Play Block (small)
				CmdLen = 0x05;
				break;
			default:	// Handle all other commands
				switch(Command & 0xF0)
				{
				case 0x30:
				case 0x40:
					CmdLen = 0x02;
					if (! ChipCommandIsValid(Command))
						printf("Warning! Command %02X found, but associated chip not used!", Command);
					break;
				case 0x50:
				case 0xA0:
				case 0xB0:
					CmdLen = 0x03;
					if (! ChipCommandIsValid(Command))
						printf("Warning! Command %02X found, but associated chip not used!", Command);
					break;
				case 0xC0:
				case 0xD0:
					CmdLen = 0x04;
					if (! ChipCommandIsValid(Command))
						printf("Warning! Command %02X found, but associated chip not used!", Command);
					break;
				case 0xE0:
				case 0xF0:
					CmdLen = 0x05;
					if (! ChipCommandIsValid(Command))
						printf("Warning! Command %02X found, but associated chip not used!", Command);
					break;
				default:
					CmdLen = 0x01;
					//StopVGM = true;
					if (! BadCmdFound)
					{
						BadCmdFound = true;
						printf("Warning! Command with unknown argument length found!\n");
					}
					break;
				}
				break;
			}
		}
		CurPos += CmdLen;
		CountLen += CmdDelay;
		if (InLoop)
			CountLoop += CmdDelay;
		
		if (StopVGM)
			break;
	}
	VGMEoDPos = CurPos;
	
	// VGMErr Bits:
	//	Bit	Val	Error Description
	//	 0	 01	Total Samples
	//	 1	 02	Loop Samples / Loop Offset
	//	 2	 04	Bad Data Offset
	//	 4	 10	GD3 Tag Offset
	//	 5	 20	GD3 Tag Length / EOF Offset
	// GD3Flags Bits:
	//	Bit	Val	Error Description
	//	0-3	 0F	Checked
	//	4-7	 F0	In Header
	//	 0	 11	GD3 found
	//	 1	 02	GD3 Offset wrong
	//	 3	 04	
	//	 4	 08	
	//VGMErr = 0x00;
	GD3Flags = 0x00;
	if (CountLen != VGMHead.lngTotalSamples)
		VGMErr |= 0x01;
	if (CountLoop != VGMHead.lngLoopSamples)
		VGMErr |= 0x02;
	
	if (CurPos > 0x04 + VGMHead.lngEOFOffset)
		printf("Warning! EOF Offset before EOF command!\n");
	if (! StopVGM)
		printf("Warning! File end before EOF command!\n");
	if (HasLoop && ! InLoop)
		printf("Warning! Loop offset between commands!\n");
	
	if (VGMHead.lngGD3Offset)
	{
		// Check for GD3 Tag at GD3 Offset
		CurPos = 0x14 + VGMHead.lngGD3Offset;
		if (CurPos <= VGMDataLen - 0x04)	// aviod crashes
			memcpy(&TempLng, &VGMData[CurPos], 0x04);
		else
			TempLng = 0x00;
		if (TempLng == FCC_GD3)
		{
			GD3Flags |= 0x01;	// GD3 Check - found
			VGMGD3Pos = CurPos;
		}
	}
	if (! (GD3Flags & 0x01))
	{
		if (VGMEoDPos <= VGMDataLen - 0x04)
		{
			// Check for GD3 Tag at EoD
			CurPos = VGMEoDPos;
			memcpy(&TempLng, &VGMData[CurPos], 0x04);
		}
		else
		{
			TempLng = 0x00;
		}
		if (TempLng != FCC_GD3)
		{
			GD3Flags |= 0x00;	// GD3 Check - missing
			if (VGMHead.lngGD3Offset)
			{
				printf("GD3 Offset set, but GD3 Tag not found!\n");
				VGMErr |= 0x10;
				GD3Flags |= 0x10;	// GD3 Header - set
			}
			//else - all ok
		}
		else
		{
			GD3Flags |= 0x01;	// GD3 Check - found
			VGMGD3Pos = CurPos;
		}
	}
	if (GD3Flags & 0x01)
	{
		CurPos = VGMGD3Pos;
		if (! VGMHead.lngGD3Offset)
		{
			printf("GD3 Tag found, but GD3 Offset missing!\n");
			VGMErr |= 0x10;
			GD3Flags |= 0x00;	// GD3 Header - not set
		}
		else if (CurPos != 0x14 + VGMHead.lngGD3Offset)	// CurPos == GD3 Offset ?
		{
			printf("Wrong GD3 Tag Offset! (Header: 0x%06lX  File: 0x%06lX)\n",
					0x14 + VGMHead.lngGD3Offset, CurPos);
			VGMErr |= 0x10;
			GD3Flags |= 0x02;	// GD3 Offset - wrong
		}
		
		memcpy(&GD3Ver, &VGMData[CurPos + 0x04], 0x04);
		if (GD3Ver > 0x00000100)
			printf("GD3 Tag version newer than supported - correction may be skipped!\n");
		
		memcpy(&CmdLen, &VGMData[CurPos + 0x08], 0x04);
		TempLng = CalcGD3Length(CurPos + 0x0C, GD3T_ENT_V100);
		if (TempLng != CmdLen)
		{
			printf("Wrong GD3 Length! (Header: 0x%06lX  File: 0x%06lX)\n", CmdLen, TempLng);
			if (TempLng < CmdLen && GD3Ver > 0x00000100)
				TempLng = CmdLen;
			else
				VGMErr |= 0x20;
		}
		
		TempLng += 0x0C;
		if (CurPos + TempLng != 0x04 + VGMHead.lngEOFOffset)	// Check EOF Offset
		{
			printf("Wrong EOF Offset! (Header: 0x%06lX  File: 0x%06lX)\n",
					0x04 + VGMHead.lngEOFOffset, CurPos + TempLng);
			VGMErr |= 0x20;
		}
	}
	else
	{
		if (VGMEoDPos != 0x04 + VGMHead.lngEOFOffset)	// Check EOF Offset
		{
			printf("Wrong EOF Offset! (Header: 0x%06lX  File: 0x%06lX)\n",
					0x04 + VGMHead.lngEOFOffset, VGMEoDPos);
			VGMErr |= 0x20;
		}
	}
	
	printf("\t%9s  %9s\n", "Header", "Counted");
	printf("Length\t%9lu  %9lu\n", VGMHead.lngTotalSamples, CountLen);
	printf("Loop\t%9lu  %9lu\n", VGMHead.lngLoopSamples, CountLoop);
	
	if (! VGMErr)
		return 0x00;	// No errors - It's all okay
	
	if (BadCmdFound && VGMHead.lngVersion > 0x00000160)
	{
		printf("Unknown Commands found and VGM version newer than supported!\n"
				"Set to Read-Only Mode!\n");
		Mode = 0x00;	// Set Read-Only-Mode
	}
	
	printf("There are some errors. ");
	if ((Mode & 0x03) == 0x00)
	{
		printf("Please try -Check to fix them.\n");
	}
	else if ((Mode & 0x03) == 0x03)
	{
		printf("Press ESC / F (fix)");
		if (VGMErr & 0x02)
			printf(" / R (fix by relocating)");
		printf(": ");
#ifdef WIN32
		TempByt = 0x00;
		while(! (TempByt == 0x1B || TempByt == 'F' || TempByt == 'R'))
		{
			TempByt = (unsigned char)toupper(_getch());
		}
		if (TempByt == 0x1B)
			printf("ESC\n");
		else
			printf("%c\n", TempByt);
#else
		TempByt = (unsigned char)toupper(getchar());
#endif
		if (TempByt == 'F')
			Mode = 0x01;
		else if (TempByt == 'R')
			Mode = 0x02;
		else
			Mode = 0x00;
	}
	else
	{
		printf("\n");
	}
	
	if (! (Mode & 0x03))
		return 0x00;
	
	RetVal = 0x00;
	if (VGMErr & 0x01)
	{
		VGMHead.lngTotalSamples = CountLen;
		RetVal |= 0x01;
	}
	//if (HasLoop && CountLoop != VGMHead.lngLoopSamples)
	if (HasLoop && VGMErr & 0x02)
	{
		if (Mode == 0x01)
		{
			if (InLoop)
			{
				VGMHead.lngLoopSamples = CountLoop;
				RetVal |= 0x02;
			}
			else
			{
				printf("Can't fix loop samples due to incorrect loop offset!\n");
				printf("Try to relocate the loop offset!\n");
			}
		}
		else if (Mode == 0x02)
		{
			LoopPos = RelocateVGMLoop();
			if (LoopPos)
			{
				VGMHead.lngLoopOffset = LoopPos - 0x0000001C;
				RetVal |= 0x02;
			}
			else
			{
				printf("Failed to relocate loop offset!\n");
			}
		}
	}
	if (VGMErr & 0x04)
	{
		VGMHead.lngDataOffset = NewDataPos - 0x00000034;
		RetVal |= 0x01;
	}
	
	//	 4	 10	GD3 Tag Offset
	//	 5	 20	GD3 Tag Length / EOF Offset
	if (VGMErr & 0x10)
	{
		switch(GD3Flags & 0x11)
		{
		case 0x10:	// GD3 offset set, but no GD3 found
			VGMHead.lngGD3Offset = 0x00000000;
			break;
		case 0x01:	// GD3 offset missing, but GD3 found
			GD3Flags |= 0x02;	// GD3 offset correction
			break;
		case 0x00:	// GD3 completely missing
		case 0x11:	// GD3 found, offset is good
			break;
		}
		
		if (GD3Flags & 0x02)	// GD3 Offset wrong
		{
			VGMHead.lngGD3Offset = VGMGD3Pos - 0x14;
		}
		RetVal |= 0x10;
	}
	if (VGMErr & 0x20)
	{
		if (GD3Flags & 0x01)
		{
			CurPos = 0x14 + VGMHead.lngGD3Offset;
			memcpy(&CmdLen, &VGMData[CurPos + 0x08], 0x04);
			TempLng = CalcGD3Length(CurPos + 0x0C, GD3T_ENT_V100);
			if (TempLng != CmdLen)
			{
				if (TempLng < CmdLen && GD3Ver > 0x00000100)
				{
					printf("GD3 Tag longer than counted - skip correction!\n");
					TempLng = CmdLen;
				}
				memcpy(&VGMData[CurPos + 0x08], &TempLng, 0x04);	// rewrite GD3 length
			}
			
			TempLng += 0x0C;
			VGMEoDPos = CurPos + TempLng;
		}
		VGMDataLen = VGMEoDPos;
		VGMHead.lngEOFOffset = VGMDataLen - 0x04;
		RetVal |= 0x20;
	}
	
	return RetVal;
}

static bool ChipCommandIsUnknown(unsigned char Command)
{
	return ((Command < 0x4F && Command != 0x30) || /*(Command >= 0xA1 && Command <= 0xAF) ||*/
			(Command >= 0xB3 && Command <= 0xBF) || (Command >= 0xC3 && Command <= 0xCF) ||
			(Command >= 0xD2 && Command <= 0xDF) || Command >= 0xE1);
}

static bool ChipCommandIsValid(unsigned char Command)
{
	if (ChipCommandIsUnknown(Command))
		return false;
	
	if (Command >= 0x61 && Command <= 0x63)
		return true;
	if (Command == 0x67)
		return true;
	if (((Command == 0x50 || Command == 0x4F) && VGMHead.lngHzPSG) ||
		((Command == 0x30 || Command == 0x3F) && (VGMHead.lngHzPSG & 0x40000000)))
		return true;
	if ((Command == 0x51 && VGMHead.lngHzYM2413) ||
		(Command == 0xA1 && (VGMHead.lngHzYM2413 & 0x40000000)))
		return true;
	if (((Command == 0x52 || Command == 0x53) && VGMHead.lngHzYM2612) ||
		((Command == 0xA2 || Command == 0xA3) && (VGMHead.lngHzYM2612 & 0x40000000)))
		return true;
	if ((Command == 0x54 && VGMHead.lngHzYM2151) ||
		(Command == 0xA4 && (VGMHead.lngHzYM2151 & 0x40000000)))
		return true;
	if (Command == 0xC0 && VGMHead.lngHzSPCM)
		return true;
	if ((Command == 0xB0 || Command == 0xC1) && VGMHead.lngHzRF5C68)
		return true;
	if ((Command == 0x55 && VGMHead.lngHzYM2203) ||
		(Command == 0xA5 && (VGMHead.lngHzYM2203 & 0x40000000)))
		return true;
	if (((Command == 0x56 || Command == 0x57) && VGMHead.lngHzYM2608) ||
		((Command == 0xA6 || Command == 0xA7) && (VGMHead.lngHzYM2608 & 0x40000000)))
		return true;
	if (((Command == 0x58 || Command == 0x59) && VGMHead.lngHzYM2610) ||
		((Command == 0xA8 || Command == 0xA9) && (VGMHead.lngHzYM2610 & 0x40000000)))
		return true;
	if ((Command == 0x5A && VGMHead.lngHzYM3812) ||
		(Command == 0xAA && (VGMHead.lngHzYM3812 & 0x40000000)))
		return true;
	if ((Command == 0x5B && VGMHead.lngHzYM3526) ||
		(Command == 0xAB && (VGMHead.lngHzYM3526 & 0x40000000)))
		return true;
	if ((Command == 0x5C && VGMHead.lngHzY8950) ||
		(Command == 0xAC && (VGMHead.lngHzY8950 & 0x40000000)))
		return true;
	if (((Command == 0x5E || Command == 0x5F) && VGMHead.lngHzYMF262) ||
		((Command == 0xAE || Command == 0xAF) && (VGMHead.lngHzYMF262 & 0x40000000)))
		return true;
	if (Command == 0xD0 && VGMHead.lngHzYMF278B)
		return true;
	if (Command == 0xD1 && VGMHead.lngHzYMF271)
		return true;
	if ((Command == 0x5D && VGMHead.lngHzYMZ280B) ||
		(Command == 0xAD && (VGMHead.lngHzYMZ280B & 0x40000000)))
		return true;
	if ((Command == 0xB1 || Command == 0xC2) && VGMHead.lngHzRF5C164)
		return true;
	if (Command == 0xB2 && VGMHead.lngHzPWM)
		return true;
	if (Command == 0xA0 && VGMHead.lngHzAY8910)
		return true;
	if (Command == 0x68 && (VGMHead.lngHzRF5C68 || VGMHead.lngHzRF5C164))
		return true;
	
	return false;
}

static unsigned long int RelocateVGMLoop(void)
{
	unsigned long int CurPos;
	unsigned char Command;
	unsigned char TempByt;
	//unsigned short int TempSht;
	unsigned long int TempLng;
	unsigned long int CmdLen;
	bool StopVGM;
	unsigned long int VGMSmplPos;
	unsigned long int VGMLoop;
	unsigned short int CmdDelay;
	
	VGMLoop = VGMHead.lngTotalSamples - VGMHead.lngLoopSamples;
	VGMSmplPos = 0x00;
	
	StopVGM = false;
	if (VGMHead.lngVersion < 0x00000150)
		CurPos = 0x40;
	else
		CurPos = 0x00000034 + VGMHead.lngDataOffset;
	while(CurPos < VGMDataLen)	// I have my reasons for NOT using VGMHead.lngEOFOffset
	{
		CmdLen = 0x00;
		CmdDelay = 0x00;
		Command = VGMData[CurPos + 0x00];
		if (Command >= 0x70 && Command <= 0x8F)
		{
			switch(Command & 0xF0)
			{
			case 0x70:
				CmdDelay = (Command & 0x0F) + 0x01;
				break;
			case 0x80:
				CmdDelay = Command & 0x0F;
				break;
			}
			CmdLen = 0x01;
		}
		else
		{
			switch(Command)
			{
			case 0x66:	// End Of File
				CmdLen = 0x01;
				StopVGM = true;
				break;
			case 0x62:	// 1/60s delay
				CmdDelay = 735;
				CmdLen = 0x01;
				break;
			case 0x63:	// 1/50s delay
				CmdDelay = 882;
				CmdLen = 0x01;
				break;
			case 0x61:	// xx Sample Delay
				memcpy(&CmdDelay, &VGMData[CurPos + 0x01], 0x02);
				CmdLen = 0x03;
				break;
			case 0x50:	// SN76496 write
				CmdLen = 0x02;
				break;
			case 0x67:	// Data Block (PCM Data Stream)
				TempByt = VGMData[CurPos + 0x02];
				memcpy(&TempLng, &VGMData[CurPos + 0x03], 0x04);
				CmdLen = 0x07 + TempLng;
				break;
			case 0x68:	// PCM RAM write
				CmdLen = 0x0C;
				break;
			case 0x90:	// DAC Ctrl: Setup Chip
				CmdLen = 0x05;
				break;
			case 0x91:	// DAC Ctrl: Set Data
				CmdLen = 0x05;
				break;
			case 0x92:	// DAC Ctrl: Set Freq
				CmdLen = 0x06;
				break;
			case 0x93:	// DAC Ctrl: Play from Start Pos
				CmdLen = 0x0B;
				break;
			case 0x94:	// DAC Ctrl: Stop immediately
				CmdLen = 0x02;
				break;
			case 0x95:	// DAC Ctrl: Play Block (small)
				CmdLen = 0x05;
				break;
			default:	// Handle all other commands
				switch(Command & 0xF0)
				{
				case 0x30:
				case 0x40:
					CmdLen = 0x02;
					break;
				case 0x50:
				case 0xA0:
				case 0xB0:
					CmdLen = 0x03;
					break;
				case 0xC0:
				case 0xD0:
					CmdLen = 0x04;
					break;
				case 0xE0:
				case 0xF0:
					CmdLen = 0x05;
					break;
				default:
					CmdLen = 0x01;
					//StopVGM = true;
					break;
				}
				break;
			}
		}
		CurPos += CmdLen;
		VGMSmplPos += CmdDelay;
		if (VGMSmplPos == VGMLoop)
			return CurPos;
		
		if (StopVGM)
			break;
	}
	
	return 0x00000000;
}

static unsigned long int CalcGD3Length(unsigned long int DataPos, unsigned short int TagEntries)
{
	unsigned long int CurPos;
	unsigned short int* CurChr;
	unsigned short int CurEnt;
	
	CurPos = DataPos;
	for (CurEnt = 0x00; CurEnt < TagEntries; CurEnt ++)
	{
		do
		{
			CurChr = (unsigned short int*)&VGMData[CurPos];
			CurPos += 0x02;
		} while(*CurChr);
	}
	
	return CurPos - DataPos;
}

static void StripVGMData(void)
{
	unsigned long int VGMPos;
	unsigned char* DstData;
	unsigned long int DstPos;
	unsigned long int CmdTimer;
	unsigned char ChipID;
	unsigned char Command;
	unsigned long int CmdDelay;
	unsigned long int AllDelay;
	unsigned char TempByt;
	unsigned short int TempSht;
	unsigned long int TempLng;
	unsigned long int LoopOfs;
	//unsigned long int DataLen;
	unsigned long int CmdLen;
	bool StopVGM;
	bool WriteEvent;
	unsigned long int NewLoopS;
	
	DstData = (unsigned char*)malloc(VGMDataLen + 0x100);
	AllDelay = 0;
	if (VGMHead.lngDataOffset)
		VGMPos = 0x34 + VGMHead.lngDataOffset;
	else
		VGMPos = 0x40;
	DstPos = VGMPos;
	LoopOfs = VGMHead.lngLoopOffset ? (0x1C + VGMHead.lngLoopOffset) : 0x00;
	NewLoopS = 0x00;
	memcpy(DstData, VGMData, VGMPos);	// Copy Header
	
	CmdTimer = 0;
	InitAllChips();
	StopVGM = false;
	while(VGMPos < VGMDataLen)
	{
		CmdDelay = 0;
		CmdLen = 0x00;
		Command = VGMData[VGMPos + 0x00];
		WriteEvent = true;
		
		if (Command >= 0x70 && Command <= 0x8F)
		{
			switch(Command & 0xF0)
			{
			case 0x70:
				TempSht = (Command & 0x0F) + 0x01;
				CmdDelay = TempSht;
				WriteEvent = false;
				break;
			case 0x80:
				TempSht = Command & 0x0F;
				//CmdDelay = TempSht;
				break;
			}
			CmdLen = 0x01;
		}
		else
		{
			// Cheat Mode (to use 2 instances of 1 chip)
			ChipID = 0x00;
			switch(Command)
			{
			case 0x30:
				if (VGMHead.lngHzPSG & 0x40000000)
				{
					Command += 0x20;
					ChipID = 0x01;
				}
				break;
			case 0x3F:
				if (VGMHead.lngHzPSG & 0x40000000)
				{
					Command += 0x10;
					ChipID = 0x01;
				}
				break;
			case 0xA1:
				if (VGMHead.lngHzYM2413 & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			case 0xA2:
			case 0xA3:
				if (VGMHead.lngHzYM2612 & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			case 0xA4:
				if (VGMHead.lngHzYM2151 & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			case 0xA5:
				if (VGMHead.lngHzYM2203 & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			case 0xA6:
			case 0xA7:
				if (VGMHead.lngHzYM2608 & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			case 0xA8:
			case 0xA9:
				if (VGMHead.lngHzYM2610 & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			case 0xAA:
				if (VGMHead.lngHzYM3812 & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			case 0xAB:
				if (VGMHead.lngHzYM3526 & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			case 0xAC:
				if (VGMHead.lngHzY8950 & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			case 0xAE:
			case 0xAF:
				if (VGMHead.lngHzYMF262 & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			case 0xAD:
				if (VGMHead.lngHzYMZ280B & 0x40000000)
				{
					Command -= 0x50;
					ChipID = 0x01;
				}
				break;
			}
			
			switch(Command)
			{
			case 0x66:	// End Of File
				CmdLen = 0x01;
				StopVGM = true;
				break;
			case 0x62:	// 1/60s delay
				TempSht = 735;
				CmdDelay = TempSht;
				CmdLen = 0x01;
				WriteEvent = false;
				break;
			case 0x63:	// 1/50s delay
				TempSht = 882;
				CmdDelay = TempSht;
				CmdLen = 0x01;
				WriteEvent = false;
				break;
			case 0x61:	// xx Sample Delay
				memcpy(&TempSht, &VGMData[VGMPos + 0x01], 0x02);
				CmdDelay = TempSht;
				CmdLen = 0x03;
				WriteEvent = false;
				break;
			case 0x50:	// SN76496 write
				WriteEvent = sn76496_write(VGMData[VGMPos + 0x01]);
				CmdLen = 0x02;
				break;
			case 0x51:	// YM2413 write
				WriteEvent = ym2413_write(VGMData[VGMPos + 0x01], VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0x52:	// YM2612 write port 0
			case 0x53:	// YM2612 write port 1
				TempByt = Command & 0x01;
				WriteEvent = ym2612_write(TempByt, VGMData[VGMPos + 0x01],
											VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0x67:	// PCM Data Stream
				TempByt = VGMData[VGMPos + 0x02];
				memcpy(&TempLng, &VGMData[VGMPos + 0x03], 0x04);
				
				switch(TempByt & 0xC0)
				{
				case 0x00:	// Database Block
				case 0x40:
					switch(TempByt)
					{
					case 0x00:	// YM2612 PCM Data
						if (StripVGM.YM2612.All || (StripVGM.YM2612.ChnMask & (0x01 << 6)))
							WriteEvent = false;
						break;
					case 0x01:	// RF5C68 PCM Database
						if (StripVGM.RF5C68.All)
							WriteEvent = false;
						break;
					case 0x02:	// RF5C164 PCM Database
						if (StripVGM.RF5C164.All)
							WriteEvent = false;
						break;
					default:
						if (StripVGM.Unknown)
							WriteEvent = false;
						break;
					}
					break;
				case 0x80:	// ROM/RAM Dump
					switch(TempByt)
					{
					case 0x80:	// SegaPCM ROM
						if (StripVGM.SegaPCM.All)
							WriteEvent = false;
						break;
					case 0x81:	// YM2608 DELTA-T ROM Image
						if (StripVGM.YM2608.All)
							WriteEvent = false;
						break;
					case 0x82:	// YM2610 ADPCM ROM Image
					case 0x83:	// YM2610 DELTA-T ROM Image
						if (StripVGM.YM2610.All)
							WriteEvent = false;
						break;
					case 0x84:	// YMF278B ROM Image
					case 0x87:	// YMF278B RAM Image
						if (StripVGM.YMF278B_All || StripVGM.YMF278B_WT.All)
							WriteEvent = false;
						break;
					case 0x85:	// YMF271 ROM Image
						if (StripVGM.YMF271.All)
							WriteEvent = false;
						break;
					case 0x86:	// YMZ280B ROM Image
						if (StripVGM.YMZ280B.All)
							WriteEvent = false;
						break;
					}
					break;
				case 0xC0:	// RAM Write
					switch(TempByt)
					{
					case 0xC0:	// RF5C68 RAM Database
						if (StripVGM.RF5C68.All)
							WriteEvent = false;
						break;
					case 0xC1:	// RF5C164 RAM Database
						if (StripVGM.RF5C164.All)
							WriteEvent = false;
						break;
					}
					break;
				}
				CmdLen = 0x07 + TempLng;
				break;
			case 0xE0:	// Seek to PCM Data Bank Pos
				memcpy(&TempLng, &VGMData[VGMPos + 0x01], 0x04);
				CmdLen = 0x05;
				break;
			case 0x4F:	// GG Stereo
				WriteEvent = GGStereo(VGMData[VGMPos + 0x01]);
				CmdLen = 0x02;
				break;
			case 0x54:	// YM2151 write
				WriteEvent = ym2151_write(VGMData[VGMPos + 0x01], VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0xC0:	// Sega PCM memory write
				memcpy(&TempSht, &VGMData[VGMPos + 0x01], 0x02);
				WriteEvent = segapcm_mem_write(TempSht, VGMData[VGMPos + 0x03]);
				CmdLen = 0x04;
				break;
			case 0xB0:	// RF5C68 register write
				WriteEvent = rf5c68_reg_write(VGMData[VGMPos + 0x01], VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0xC1:	// RF5C68 memory write
				memcpy(&TempSht, &VGMData[VGMPos + 0x01], 0x02);
				WriteEvent = rf5c68_mem_write(TempSht, VGMData[VGMPos + 0x03]);
				CmdLen = 0x04;
				break;
			case 0x55:	// YM2203
				WriteEvent = ym2203_write(VGMData[VGMPos + 0x01], VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0x56:	// YM2608 write port 0
			case 0x57:	// YM2608 write port 1
				TempByt = Command & 0x01;
				WriteEvent = ym2608_write(TempByt, VGMData[VGMPos + 0x01],
											VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0x58:	// YM2610 write port 0
			case 0x59:	// YM2610 write port 1
				TempByt = Command & 0x01;
				WriteEvent = ym2610_write(TempByt, VGMData[VGMPos + 0x01],
											VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0x5A:	// YM3812 write
				WriteEvent = ym3812_write(VGMData[VGMPos + 0x01], VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0x5B:	// YM3526 write
				WriteEvent = ym3526_write(VGMData[VGMPos + 0x01], VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0x5C:	// Y8950 write
				WriteEvent = y8950_write(VGMData[VGMPos + 0x01], VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0x5E:	// YMF262 write port 0
			case 0x5F:	// YMF262 write port 1
				TempByt = Command & 0x01;
				WriteEvent = ymf262_write(TempByt, VGMData[VGMPos + 0x01],
											VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0x5D:	// YMZ280B write
				WriteEvent = ymz280b_write(VGMData[VGMPos + 0x01], VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0xD0:	// YMF278B register write
				WriteEvent = true;
				if (StripVGM.YMF278B_All)
					WriteEvent = false;
				if (StripVGM.YMF278B_FM.All && VGMData[VGMPos + 0x01] < 0x02 &&
					! (VGMData[VGMPos + 0x01] == 0x01 && VGMData[VGMPos + 0x02] == 0x05))
					WriteEvent = false;	// Don't kill the WaveTable-On Command at Reg 0x105
				if (StripVGM.YMF278B_WT.All && VGMData[VGMPos + 0x01] == 0x02)
					WriteEvent = false;
				CmdLen = 0x03;
				break;
			case 0xD1:	// YMF271 register write
				WriteEvent = true;
				if (StripVGM.YMF271.All)
					WriteEvent = false;
				CmdLen = 0x03;
				break;
			case 0xB1:	// RF5C164 register write
				WriteEvent = rf5c164_reg_write(VGMData[VGMPos + 0x01], VGMData[VGMPos + 0x02]);
				CmdLen = 0x03;
				break;
			case 0xC2:	// RF5C164 memory write
				memcpy(&TempSht, &VGMData[VGMPos + 0x01], 0x02);
				WriteEvent = rf5c164_mem_write(TempSht, VGMData[VGMPos + 0x03]);
				CmdLen = 0x04;
				break;
			case 0xB2:	// PWM register write
				if (StripVGM.PWM.All)
					WriteEvent = false;
				CmdLen = 0x03;
				break;
			case 0x68:	// PCM RAM write
				TempByt = VGMData[VGMPos + 0x02];
				switch(TempByt)
				{
				case 0xC0:	// RF5C68 RAM Database
					if (StripVGM.RF5C68.All)
						WriteEvent = false;
					break;
				case 0xC1:	// RF5C164 RAM Database
					if (StripVGM.RF5C164.All)
						WriteEvent = false;
					break;
				}
				CmdLen = 0x0C;
				break;
			case 0xA0:	// AY8910 register write
				if (StripVGM.AY8910.All)
					WriteEvent = false;
				CmdLen = 0x03;
				break;
			case 0x90:	// DAC Ctrl: Setup Chip
				CmdLen = 0x05;
				break;
			case 0x91:	// DAC Ctrl: Set Data
				CmdLen = 0x05;
				break;
			case 0x92:	// DAC Ctrl: Set Freq
				CmdLen = 0x06;
				break;
			case 0x93:	// DAC Ctrl: Play from Start Pos
				CmdLen = 0x0B;
				break;
			case 0x94:	// DAC Ctrl: Stop immediately
				CmdLen = 0x02;
				break;
			case 0x95:	// DAC Ctrl: Play Block (small)
				CmdLen = 0x05;
				break;
			default:
				if (StripVGM.Unknown)
					WriteEvent = false;
				switch(Command & 0xF0)
				{
				case 0x30:
				case 0x40:
					CmdLen = 0x02;
					break;
				case 0x50:
				case 0xA0:
				case 0xB0:
					CmdLen = 0x03;
					break;
				case 0xC0:
				case 0xD0:
					CmdLen = 0x04;
					break;
				case 0xE0:
				case 0xF0:
					CmdLen = 0x05;
					break;
				default:
					printf("Unknown Command: %02hX\n", Command);
					CmdLen = 0x01;
					//StopVGM = true;
					break;
				}
				break;
			}
		}
		
		if (WriteEvent || VGMPos == LoopOfs)
		{
			if (VGMPos != LoopOfs)
			{
				AllDelay += CmdDelay;
				CmdDelay = 0x00;
			}
			while(AllDelay)
			{
				if (AllDelay <= 0xFFFF)
					TempSht = (unsigned short int)AllDelay;
				else
					TempSht = 0xFFFF;
				
				if (! TempSht)
				{
					// don't do anything - I just want to be safe
				}
				if (TempSht <= 0x10)
				{
					DstData[DstPos] = 0x70 | (TempSht - 0x01);
					DstPos ++;
				}
				else if (TempSht <= 0x20)
				{
					DstData[DstPos] = 0x7F;
					DstPos ++;
					DstData[DstPos] = 0x70 | (TempSht - 0x11);
					DstPos ++;
				}
				else if ((TempSht >=  735 && TempSht <=  751) || TempSht == 1470)
				{
					TempLng = TempSht;
					while(TempLng >= 735)
					{
						DstData[DstPos] = 0x62;
						DstPos ++;
						TempLng -= 735;
					}
					TempSht -= (unsigned short int)TempLng;
				}
				else if ((TempSht >=  882 && TempSht <=  898) || TempSht == 1764)
				{
					TempLng = TempSht;
					while(TempLng >= 882)
					{
						DstData[DstPos] = 0x63;
						DstPos ++;
						TempLng -= 882;
					}
					TempSht -= (unsigned short int)TempLng;
				}
				else if (TempSht == 1617)
				{
					DstData[DstPos] = 0x63;
					DstPos ++;
					DstData[DstPos] = 0x62;
					DstPos ++;
				}
				else
				{
					DstData[DstPos + 0x00] = 0x61;
					memcpy(&DstData[DstPos + 0x01], &TempSht, 0x02);
					DstPos += 0x03;
				}
				AllDelay -= TempSht;
			}
			AllDelay = CmdDelay;
			CmdDelay = 0x00;
			
			if (VGMPos == LoopOfs)
				NewLoopS = DstPos;
			
			if (WriteEvent)
			{
				memcpy(&DstData[DstPos], &VGMData[VGMPos], CmdLen);
				DstPos += CmdLen;
			}
		}
		else
		{
			AllDelay += CmdDelay;
		}
		VGMPos += CmdLen;
		if (StopVGM)
			break;
	}
	if (LoopOfs)
	{
		if (! NewLoopS)
		{
			printf("Error! Failed to relocate Loop Point!\n");
			NewLoopS = 0x1C;
		}
		VGMHead.lngLoopOffset = NewLoopS - 0x1C;
	}
	printf("\t\t\t\t\t\t\t\t\r");
	
	if (VGMHead.lngGD3Offset)
	{
		VGMPos = 0x14 + VGMHead.lngGD3Offset;
		memcpy(&TempLng, &VGMData[VGMPos + 0x00], 0x04);
		if (TempLng == FCC_GD3)
		{
			memcpy(&CmdLen, &VGMData[VGMPos + 0x08], 0x04);
			CmdLen += 0x0C;
			
			VGMHead.lngGD3Offset = DstPos - 0x14;
			memcpy(&DstData[DstPos], &VGMData[VGMPos], CmdLen);
			DstPos += CmdLen;
		}
	}
	VGMDataLen = DstPos;
	VGMHead.lngEOFOffset = VGMDataLen - 0x04;
	
	if (StripVGM.SN76496.All)
	{
		VGMHead.lngHzPSG = 0;
		VGMHead.shtPSG_Feedback = 0x00;
		VGMHead.bytPSG_SRWidth = 0x00;
		VGMHead.bytPSG_Flags = 0x00;
	}
	if (StripVGM.YM2413.All)
		VGMHead.lngHzYM2413 = 0;
	if (StripVGM.YM2612.All)
		VGMHead.lngHzYM2612 = 0;
	if (StripVGM.YM2151.All)
		VGMHead.lngHzYM2151 = 0;
	if (StripVGM.SegaPCM.All)
	{
		VGMHead.lngHzSPCM = 0;
		VGMHead.lngSPCMIntf = 0x00;
	}
	if (StripVGM.RF5C68.All)
		VGMHead.lngHzRF5C68 = 0;
	if (StripVGM.YM2203.All)
	{
		VGMHead.lngHzYM2203 = 0;
		VGMHead.bytAYFlagYM2203 = 0x00;
	}
	if (StripVGM.YM2608.All)
	{
		VGMHead.lngHzYM2608 = 0;
		VGMHead.bytAYFlagYM2608 = 0x00;
	}
	if (StripVGM.YM2610.All)
		VGMHead.lngHzYM2610 = 0;
	if (StripVGM.YM3812.All)
		VGMHead.lngHzYM3812 = 0;
	if (StripVGM.YM3526.All)
		VGMHead.lngHzYM3526 = 0;
	if (StripVGM.Y8950.All)
		VGMHead.lngHzY8950 = 0;
	if (StripVGM.YMF262.All)
		VGMHead.lngHzYMF262 = 0;
	if (StripVGM.YMF278B_All)
		VGMHead.lngHzYMF278B = 0;
	if (StripVGM.YMF271.All)
		VGMHead.lngHzYMF271 = 0;
	if (StripVGM.YMZ280B.All)
		VGMHead.lngHzYMZ280B = 0;
	if (StripVGM.RF5C164.All)
		VGMHead.lngHzRF5C164 = 0;
	if (StripVGM.PWM.All)
		VGMHead.lngHzPWM = 0;
	if (StripVGM.AY8910.All)
	{
		VGMHead.lngHzAY8910 = 0;
		VGMHead.bytAYType = 0x00;
		VGMHead.bytAYFlag = 0x00;
	}
	
	// PatchVGM will rewrite the header later
	
	VGMData = (unsigned char*)realloc(VGMData, VGMDataLen);
	memcpy(VGMData, DstData, VGMDataLen);
	free(DstData);
	
	FreeAllChips();
	
	return;
}
